# -*- coding: utf-8 -*-

"""
This module provides a script to test the
radish step implementations
"""

import sys

import colorful
from docopt import docopt

from radish import __VERSION__
import radish.utils as utils
from radish.testing.matches import test_step_matches_configs


# use only 8 ANSI colors
# FIXME(TF): change to true colors!
colorful.use_8_ansi_colors()


def main():
    """
    Usage:
        radish-test matches <match-configs>...
            [-b=<basedir> | --basedir=<basedir>...]
            [--cover-min-percentage=<cover-min-percentage>]
            [--cover-show-missing]
        radish-test (-h | --help)
        radish-test (-v | --version)

    Arguments:
        match-configs                                  configuration files of step matcher tests

    Commands:
        matches                                        test if the step implementations actually match the expected sentences

    Options:
        -h --help                                      show this screen
        -v --version                                   show version
        -b=<basedir> --basedir=<basedir>...            set base dir from where the step.py and terrain.py will be loaded. [default: $PWD/radish]
                                                       You can specify -b|--basedir multiple times or split multiple paths with a colon (:) similar to $PATH. All files will be imported.
        --cover-min-percentage=<cover-min-percentage>  minimum percentage of step coverage for tests to pass
        --cover-show-missing                           show steps which are not tested

    (C) Copyright by Timo Furrer <tuxtimo@gmail.com>
    """

    arguments = docopt("radish-test {0}\n{1}".format(__VERSION__, main.__doc__), version=__VERSION__)

    if arguments["matches"]:
        return test_step_matches_configs(
            arguments["<match-configs>"],
            utils.flattened_basedirs(arguments["--basedir"]),
            arguments["--cover-min-percentage"],
            arguments["--cover-show-missing"],
        )


if __name__ == "__main__":
    sys.exit(main())



# -*- coding: utf-8 -*-

from radish import given, when, then


@given("the device is {is_plugged_in:is_plugged_in}")
def device_plugged_in(step, is_plugged_in):
    step.context.powered = False
    step.context.device_plugged_in = is_plugged_in


@when("I turn on the board")
def turn_on_board(step):
    step.context.powered = True


@then("I expect the device {is_registered:is_registered}")
def device_registered(step, is_registered):
    assert step.context.powered, "Board is not powered, thus device cannot be registered"
    # device is always registered when the device is plugged in
    assert step.context.device_plugged_in == is_registered, "Device is not regsitered"



# -*- coding: utf-8 -*-

from radish import given, when, then


@given("I have the numbers {numbers:Number+}")
def have_number(step, numbers):
    step.context.numbers.extend(numbers)


@given("I have the following numbers {numbers:NumberList}")
def have_number(step, numbers):
    step.context.numbers.extend(numbers)


@when("I sum them")
def sum_numbers(step):
    step.context.result = sum(step.context.numbers)


@then("I expect the result to be {result:g}")
def expect_result(step, result):
    assert step.context.result == result


@given("I have studied the questions")
def studied_questions(step):
    pass


@when("I have to answer it")
def answer_question(step):
    pass


@then('I say "{answer:AnswerChoice}"')
def say_answer(step, answer):
    assert answer == "yes"



# -*- coding: utf-8 -*-

"""
This module provides several utility functions
"""

import os
import re
import sys
import fnmatch
import traceback
import warnings
import pydoc
import itertools
import calendar
from datetime import datetime, timedelta


class Failure(object):  # pylint: disable=too-few-public-methods
    """
    Represents the fail reason for a step
    """

    def __init__(self, exception):
        """
        Initalizes the Step failure with a given Exception

        :param Exception exception: the exception shrown in the step
        """
        self.exception = exception
        self.reason = str(exception)
        self.traceback = traceback.format_exc()
        self.name = exception.__class__.__name__
        traceback_info = traceback.extract_tb(sys.exc_info()[2])[-1]
        self.filename = traceback_info[0]
        self.line = int(traceback_info[1])


def console_write(text):
    """
    Writes the given text to the console

    :param str text: the text which is printed to the console
    """
    print(str(text))


def expandpath(path):
    """
    Expands a path

    :param string path: the path to expand
    """
    return os.path.expanduser(os.path.expandvars(path))


def recursive_glob(root, pattern):
    """
    Recursively search for files with given pattern inside a path

    :param str root: the root location to start search
    :param str pattern: to pattern to look for. It's matched against the filenames under `root`.

    :rtype: list
    :returns: A list of matching files
    """
    matches = []
    for root, dirnames, filenames in os.walk(root):
        for filename in fnmatch.filter(filenames, pattern):
            matches.append(os.path.join(root, filename))
    return matches


def get_debugger():
    """
    Returns a debugger instance
    """
    try:
        from IPython.core.debugger import Pdb

        pdb = Pdb()
    except ImportError:
        try:
            from IPython.Debugger import Pdb
            from IPython.Shell import IPShell

            IPShell(argv=[""])
            pdb = Pdb()
        except ImportError:
            warnings.warn(
                'pdb was selected as a debugger. If you want to use ipython as a debugger you have to "pip install radish-bdd[ipython-debugger]"'
            )
            import pdb

    return pdb


def format_utc_to_local_tz(utc_dt, fmt="%Y-%m-%dT%H:%M:%S"):
    """
    Formats the given UTC datetime as a string converted to the local timezone.

    If the datetime is ``None``, it will return an empty string.

    Args:
        datetime (Datetime): the UTC based datetime object
        fmt (str): optional format for the string
    """
    if not utc_dt:
        return ""

    def utc_to_local(utc_dt):
        timestamp = calendar.timegm(utc_dt.timetuple())
        local_dt = datetime.fromtimestamp(timestamp)
        assert utc_dt.resolution >= timedelta(microseconds=1)
        return local_dt.replace(microsecond=utc_dt.microsecond)

    return utc_to_local(utc_dt).strftime(fmt)


def make_unique_obj_list(somelist, attr):
    """
    Make list with objects unique
    according to an objects attribute.
    """
    tmp = {}
    for item in somelist:
        tmp[attr(item)] = item
    return tmp.values()


def get_func_code(func):
    """
    Get the code object for the given function.
    """
    if sys.version_info[0] == 3:
        return func.__code__
    else:
        return func.func_code


def get_func_arg_names(func):
    """
    Get the argument names of the given function.
    """
    return get_func_code(func).co_varnames


def get_func_location(func):
    """
    Get the location where the given function is implemented.
    """
    func_code = get_func_code(func)
    return "{0}:{1}".format(func_code.co_filename, func_code.co_firstlineno)


def str_lreplace(pattern, replacement, string, escape_pattern=False, flags=0):
    """
    Only replace the pattern with replacement
    if the string starts with the pattern.
    """
    if escape_pattern:
        pattern = re.escape(pattern)

    return re.sub(r"^{0}".format(pattern), replacement, string, flags=flags)


def locate(name):
    """
    Locate the object for the given name
    """
    obj = pydoc.locate(name)
    if not obj:
        obj = globals().get(name, None)

    return obj


def flattened_basedirs(basedirs):
    """
    Flatten a list of basedirs.

    Multiple basedirs can be specified within a
    single element split by a colon.
    """
    separator = ";" if os.name == "nt" else ":"
    return list(x for x in itertools.chain(*(x.split(separator) for x in basedirs)) if x)


def split_unescape(s, delim, escape="\\", unescape=True):
    """
    Split a string based on a delimiter and an escape character.

    :param str s: The input string to split and unescape.
    :param str delim: The delimiter used for splitting the string.
    :param str escape: The escape character used to escape special characters (default is "\\").
    :param bool unescape: A flag to control whether to unescape escaped characters (default is True).

    :rtype: list
    :returns: A list of split and unescaped strings.

    Examples:
    >>> split_unescape('foo|bar', '|')
    ['foo', 'bar']
    >>> split_unescape('foo\\|bar', '|')
    ['foo|bar']
    >>> split_unescape(r'foo\\|bar', '|', unescape=True)
    ['foo\\', 'bar']
    >>> split_unescape(r'foo\\|bar', '|', unescape=False)
    ['foo\\\\', 'bar']
    >>> split_unescape('foo\\', '|', unescape=True)
    ['foo\\']
    """
    ret = []
    current = []
    itr = iter(s)
    for ch in itr:
        if ch == escape:
            try:
                # skip the next character; it has been escaped!
                if not unescape:
                    current.append(escape)
                current.append(next(itr))
            except StopIteration:
                if unescape:
                    current.append(escape)
        elif ch == delim:
            # split! (add current to the list and reset it)
            ret.append("".join(current))
            current = []
        else:
            current.append(ch)
    ret.append("".join(current))
    return ret



# -*- coding: utf-8 -*-

"""
    radish
    ~~~~~~

    Behavior Driven Development tool for Python - the root from red to green

    Copyright: MIT, Timo Furrer <tuxtimo@gmail.com>
"""

import os

import pytest

from radish.terrain import world
from radish.core import Core, Configuration
from radish.parser import FeatureParser
from radish.stepregistry import StepRegistry
from radish.hookregistry import HookRegistry
from radish.extensionregistry import ExtensionRegistry

#: Holds the path to the Feature file resources
__TEST_BASE_DIR__ = os.path.dirname(__file__)
__FEATURE_FILES_DIR__ = os.path.join(__TEST_BASE_DIR__, "features")
__RADISH_FILES_DIR__ = os.path.join(__TEST_BASE_DIR__, "radish")
__OUTPUT_DIR__ = os.path.join(__TEST_BASE_DIR__, "output")


@pytest.fixture(scope="function", autouse=True)
def mock_world_config():
    """
    Fixture to mock the terrain.world.config object
    with some default fake data.
    """
    # default command line arguments
    arguments = {
        "--basedir": ["$PWD/radish"],
        "--bdd-xml": None,
        "--cover-append": False,
        "--cover-branches": False,
        "--cover-config-file": ".coveragerc",
        "--cover-erase": False,
        "--cover-html": None,
        "--cover-min-percentage": None,
        "--cover-packages": None,
        "--cover-xml": None,
        "--cucumber-json": None,
        "--debug-after-failure": False,
        "--debug-steps": False,
        "--dry-run": False,
        "--early-exit": False,
        "--expand": False,
        "--help": False,
        "--inspect-after-failure": False,
        "--junit-xml": None,
        "--junit-relaxed": False,
        "--marker": "time.time()",
        "--no-ansi": False,
        "--no-line-jump": False,
        "--profile": None,
        "--scenarios": None,
        "--shuffle": False,
        "--syslog": False,
        "--tags": None,
        "--user-data": [],
        "--version": False,
        "--with-coverage": False,
        "--with-traceback": False,
        "--write-ids": False,
        "--write-steps-once": False,
        "<features>": ["features/"],
        "show": False,
    }
    world.config = Configuration(arguments)
    yield world.config
    delattr(world, "config")


@pytest.fixture(scope="function", autouse=True)
def reset_registries():
    """
    Fixture to automatically reset singleton registries
    """
    StepRegistry().clear()
    HookRegistry().reset()
    ExtensionRegistry().reset()


@pytest.fixture
def world_config(mock_world_config):
    """
    Fixture to work with world.config object
    """
    yield mock_world_config


@pytest.fixture()
def mock_utils_debugger(mocker):
    """
    Fixture to mock the pdf Python debugger
    """

    def call_orig_func(func, *args, **kwargs):
        """
        Helper to mock pdf.runcall interface
        """
        return func(*args, **kwargs)

    debugger_mock = mocker.patch("radish.utils.get_debugger")
    debugger_mock.return_value.runcall = mocker.MagicMock(side_effect=call_orig_func)
    return debugger_mock.return_value


@pytest.fixture()
def core():
    """
    Fixture to radish.core.Core
    """
    return Core()


@pytest.fixture()
def featurefile(request):
    """
    Fixture to get the path to a Feature File
    """
    featurename = request.param[0]
    return os.path.join(__FEATURE_FILES_DIR__, featurename + ".feature")


@pytest.fixture()
def featurefiledir():
    """
    Fixture to return the location of the test feature file dir
    """
    return __FEATURE_FILES_DIR__


@pytest.fixture()
def radishdir():
    """
    Fixture to return the location of the test radish files dir
    """
    return __RADISH_FILES_DIR__


@pytest.fixture()
def outputdir():
    """
    Fixture to return the location of the test output dir
    """
    return __OUTPUT_DIR__


@pytest.fixture()
def parser(request, core):
    """
    Fixture to create a Feature Parser ready to parse the given Feature File
    """
    featurename = request.param[0]
    # get parser positional arguments
    try:
        parser_args = request.param[1]
    except IndexError:
        parser_args = []
    # get parser keyword arguments
    try:
        parser_kwargs = request.param[2]
    except IndexError:
        parser_kwargs = {}

    # create the Feature Parser instance
    return FeatureParser(
        core, os.path.join(__FEATURE_FILES_DIR__, featurename + ".feature"), 1, *parser_args, **parser_kwargs
    )


@pytest.fixture()
def stepregistry():
    """
    Fixture to create and get a clean StepRegistry instance.
    """
    registry = StepRegistry()
    yield registry


@pytest.fixture()
def hookregistry():
    """
    Fixture to create and get a clean HookRegistry instance.
    """
    registry = HookRegistry()
    yield registry


@pytest.fixture()
def extensionregistry():
    """
    Fixture to create and get a clean ExtensionRegistry instance.
    """
    registry = ExtensionRegistry()
    yield registry



# -*- coding: utf-8 -*-

"""
    radish
    ~~~~~~

    Behavior Driven Development tool for Python - the root from red to green

    Copyright: MIT, Timo Furrer <tuxtimo@gmail.com>
"""

from radish.scenarioloop import ScenarioLoop
from radish.iterationscenario import IterationScenario
from radish.background import Background
from radish.stepmodel import Step


def test_creating_simple_scenarioloop():
    """
    Test creating a simple ScenarioLoop
    """
    # given & when
    scenario = ScenarioLoop(
        1,
        "Scenario Loop",
        "Iterations",
        "I am a Scenario Loop",
        "foo.feature",
        1,
        parent=None,
        tags=None,
        preconditions=None,
        background=None,
    )

    # then
    assert scenario.id == 1
    assert scenario.keyword == "Scenario Loop"
    assert scenario.iterations_keyword == "Iterations"
    assert scenario.sentence == "I am a Scenario Loop"
    assert scenario.path == "foo.feature"
    assert scenario.line == 1
    assert scenario.parent is None
    assert scenario.tags == []
    assert scenario.preconditions == []
    assert scenario.background is None


def test_building_scenarioloop_scenarios(mocker):
    """
    Test building Scenarios from a Scenario Loop
    """
    # given
    scenario_loop = ScenarioLoop(
        1,
        "Scenario Loop",
        "Iterations",
        "I am a Scenario Loop",
        "foo.feature",
        1,
        parent=None,
        tags=None,
        preconditions=None,
        background=None,
    )
    # add steps
    scenario_loop.steps.extend(
        [
            mocker.MagicMock(sentence="Given I have 1", path="foo.feature"),
            mocker.MagicMock(sentence="And I have 2", path="foo.feature"),
            mocker.MagicMock(sentence="When I add those", path="foo.feature"),
        ]
    )
    # set iterations
    scenario_loop.iterations = 2

    # when - build the scenarios
    scenario_loop.build_scenarios()

    # then - expect 2 built Scenarios
    assert len(scenario_loop.scenarios) == 2
    # then - expect that Scenarios are of type ExampleScenario
    assert all(isinstance(x, IterationScenario) for x in scenario_loop.scenarios)
    # then - expect correct Example Scenario sentences
    assert scenario_loop.scenarios[0].sentence == "I am a Scenario Loop - iteration 0"
    assert scenario_loop.scenarios[1].sentence == "I am a Scenario Loop - iteration 1"
    # then - expect correctly replaced Step sentences
    assert scenario_loop.scenarios[0].steps[0].sentence == "Given I have 1"
    assert scenario_loop.scenarios[0].steps[1].sentence == "And I have 2"
    assert scenario_loop.scenarios[0].steps[2].sentence == "When I add those"
    assert scenario_loop.scenarios[1].steps[0].sentence == "Given I have 1"
    assert scenario_loop.scenarios[1].steps[1].sentence == "And I have 2"
    assert scenario_loop.scenarios[1].steps[2].sentence == "When I add those"


def test_building_scenarioloop_scenarios_with_background(mocker):
    """
    Test building Scenarios from a Scenario Loop including a Background
    """
    # given
    background = Background("Background", "I am a Background", "foo.feature", 1, parent=None)
    # add some Steps
    background.steps.extend(
        [
            Step(1, "Foo", "foo.feature", 2, background, False),
            Step(2, "Foo", "foo.feature", 3, background, False),
        ]
    )
    scenario_loop = ScenarioLoop(
        1,
        "Scenario Loop",
        "Iterations",
        "I am a Scenario Loop",
        "foo.feature",
        1,
        parent=None,
        tags=None,
        preconditions=None,
        background=background,
    )
    # add steps
    scenario_loop.steps.extend(
        [
            mocker.MagicMock(sentence="Given I have 1", path="foo.feature"),
            mocker.MagicMock(sentence="And I have 2", path="foo.feature"),
            mocker.MagicMock(sentence="When I add those", path="foo.feature"),
        ]
    )
    # set iterations
    scenario_loop.iterations = 2

    # when - build the scenarios
    scenario_loop.build_scenarios()

    # then - expect ExampleScenarios to have background copy assigned
    assert scenario_loop.scenarios[0].background.sentence == "I am a Background"
    assert scenario_loop.scenarios[1].background.sentence == "I am a Background"


def test_scenarioloop_afterparse_logic(mocker):
    """
    Test Scenario Loop after parse logic
    """
    # given
    scenario_loop = ScenarioLoop(
        1,
        "Scenario Loop",
        "Iterations",
        "I am a Scenario Loop",
        "foo.feature",
        1,
        parent=None,
        tags=None,
        preconditions=None,
        background=None,
    )
    # add steps
    scenario_loop.steps.extend(
        [
            mocker.MagicMock(sentence="Given I have 1", path="foo.feature"),
            mocker.MagicMock(sentence="And I have 2", path="foo.feature"),
            mocker.MagicMock(sentence="When I add those", path="foo.feature"),
        ]
    )
    # set iterations
    scenario_loop.iterations = 2

    # when
    scenario_loop.after_parse()

    # then - expect 2 built Scenarios
    assert len(scenario_loop.scenarios) == 2
    assert scenario_loop.complete is True



# -*- coding: utf-8 -*-

"""
This module provides a class to represent a Scenario Outline
"""

import copy

from .scenario import Scenario
from .examplescenario import ExampleScenario
from .stepmodel import Step
from .exceptions import RadishError


class ScenarioOutline(Scenario):
    """
    Represents a Scenario
    """

    class Example(object):
        """
        Represents the ScenarioOutline examples
        """

        def __init__(self, data, path, line):
            self.data = data
            self.path = path
            self.line = line

    def __init__(
        self,
        id,
        keyword,
        example_keyword,
        sentence,
        path,
        line,
        parent,
        tags=None,
        preconditions=None,
        background=None,
    ):
        super(ScenarioOutline, self).__init__(
            id, keyword, sentence, path, line, parent, tags, preconditions, background
        )
        self.example_keyword = example_keyword
        self.scenarios = []
        self.examples_header = []
        self.examples = []

    def build_scenarios(self):
        """
        Builds the scenarios with the parsed Examples

        Note: This must be done before merging the steps
              from the feature file with the step definitions
        """
        for row_id, example in enumerate(self.examples):
            examples = dict(zip(self.examples_header, example.data))
            scenario_id = self.id + row_id + 1
            background = None
            scenario = ExampleScenario(
                scenario_id,
                self.keyword,
                "{0} - row {1}".format(self.sentence, row_id),
                self.path,
                self.line,
                self,
                example,
            )
            if self.background:
                background = self.background.create_instance(parent=scenario, steps_runable=True)
                scenario.background = background

            for step_id, outlined_step in enumerate(self.steps):
                sentence = self._replace_examples_in_sentence(outlined_step.sentence, examples)
                raw_text = self._replace_examples_in_raw_text(outlined_step.raw_text, examples)
                step = Step(
                    step_id + 1,
                    sentence,
                    outlined_step.path,
                    example.line,
                    scenario,
                    True,
                    context_class=outlined_step.context_class,
                )
                # copy extended attributes
                step.table_header = copy.copy(outlined_step.table_header)
                step.table_data = copy.copy(outlined_step.table_data)
                step.table = copy.copy(outlined_step.table)
                step.raw_text = raw_text

                # add step to scenario
                scenario.steps.append(step)
            self.scenarios.append(scenario)

    @staticmethod
    def _replace_examples_in_sentence(sentence, examples):
        """
        Replaces the given examples in the given sentence

        :param string sentence: the step sentence in which to replace the examples
        :param dict examples: the examples

        :returns: the new step sentence
        :rtype: string
        """
        for key, value in examples.items():
            sentence = sentence.replace("<{0}>".format(key), value)
        return sentence

    @staticmethod
    def _replace_examples_in_raw_text(raw_text, examples):
        """
        Replaces the given examples in the given raw_text, which is
        a list of strings

        :param list raw_text: list of strings in which to replace the examples
        :param dict examples: the examples

        :returns: the new list raw_text
        :rtype: list
        """
        raw_text_list = []
        for raw_text_item in raw_text:
            raw_text_item = ScenarioOutline._replace_examples_in_sentence(raw_text_item, examples)
            raw_text_list.append(raw_text_item)
        return raw_text_list

    def get_column_width(self, column_index):
        """
        Gets the column width from the given column

        :param int column_index: the column index to get the width from
        """
        try:
            return max(
                max([len(x.data[column_index]) for x in self.examples]),
                len(self.examples_header[column_index]),
            )
        except IndexError:
            raise RadishError("Invalid colum_index to get column width for ScenarioOutline '{0}'".format(self.sentence))

    def after_parse(self):
        """
        Build outlined scenarios
        """
        Scenario.after_parse(self)
        self.build_scenarios()
        self.complete = True



# -*- coding: utf-8 -*-

"""
This radish extension module provide the functionality to write the end report
"""

# disable no-member lint error because of dynamic method from colorful
# pylint: disable=no-member

from datetime import timedelta

import colorful
import humanize

from radish.hookregistry import after
from radish.stepmodel import Step
from radish.utils import console_write as write, make_unique_obj_list, get_func_code
from radish.scenariooutline import ScenarioOutline
from radish.scenarioloop import ScenarioLoop
from radish.extensionregistry import extension
from radish.terrain import world
from radish.stepregistry import StepRegistry


@extension
class EndreportWriter(object):
    """
    Endreport writer radish extension
    """

    LOAD_IF = staticmethod(lambda config: not config.show)
    LOAD_PRIORITY = 50

    def __init__(self):
        after.all(self.console_write)

    def console_write(self, features, marker):
        """
        Writes the endreport for all features

        :param list features: all features
        """
        stats = {
            "features": {
                "amount": 0,
                "passed": 0,
                "failed": 0,
                "skipped": 0,
                "untested": 0,
                "pending": 0,
            },
            "scenarios": {
                "amount": 0,
                "passed": 0,
                "failed": 0,
                "skipped": 0,
                "untested": 0,
                "pending": 0,
            },
            "steps": {
                "amount": 0,
                "passed": 0,
                "failed": 0,
                "skipped": 0,
                "untested": 0,
                "pending": 0,
            },
        }
        pending_steps = []
        duration = timedelta()
        for feature in features:
            if not feature.has_to_run(world.config.scenarios):
                continue
            stats["features"]["amount"] += 1
            stats["features"][feature.state] += 1

            if feature.state in [Step.State.PASSED, Step.State.FAILED]:
                duration += feature.duration

            for scenario in feature.all_scenarios:
                if not scenario.has_to_run(world.config.scenarios):
                    continue

                if isinstance(scenario, ScenarioOutline):  # skip ScenarioOutlines
                    continue
                if isinstance(scenario, ScenarioLoop):  # skip ScenarioLoop
                    continue

                stats["scenarios"]["amount"] += 1
                stats["scenarios"][scenario.state] += 1
                for step in scenario.steps:
                    stats["steps"]["amount"] += 1
                    stats["steps"][step.state] += 1

                    if step.state == Step.State.PENDING:
                        pending_steps.append(step)

        colored_closing_paren = colorful.bold_white(")")
        colored_comma = colorful.bold_white(", ")
        passed_word = colorful.bold_green("{0} passed")
        failed_word = colorful.bold_red("{0} failed")
        skipped_word = colorful.cyan("{0} skipped")
        pending_word = colorful.bold_yellow("{0} pending")

        output = colorful.bold_white("{0} features (".format(stats["features"]["amount"]))
        output += passed_word.format(stats["features"]["passed"])
        if stats["features"]["failed"]:
            output += colored_comma + failed_word.format(stats["features"]["failed"])
        if stats["features"]["skipped"]:
            output += colored_comma + skipped_word.format(stats["features"]["skipped"])
        if stats["features"]["pending"]:
            output += colored_comma + pending_word.format(stats["features"]["pending"])
        output += colored_closing_paren

        output += "\n"
        output += colorful.bold_white("{} scenarios (".format(stats["scenarios"]["amount"]))
        output += passed_word.format(stats["scenarios"]["passed"])
        if stats["scenarios"]["failed"]:
            output += colored_comma + failed_word.format(stats["scenarios"]["failed"])
        if stats["scenarios"]["skipped"]:
            output += colored_comma + skipped_word.format(stats["scenarios"]["skipped"])
        if stats["scenarios"]["pending"]:
            output += colored_comma + pending_word.format(stats["scenarios"]["pending"])
        output += colored_closing_paren

        output += "\n"
        output += colorful.bold_white("{} steps (".format(stats["steps"]["amount"]))
        output += passed_word.format(stats["steps"]["passed"])
        if stats["steps"]["failed"]:
            output += colored_comma + failed_word.format(stats["steps"]["failed"])
        if stats["steps"]["skipped"]:
            output += colored_comma + skipped_word.format(stats["steps"]["skipped"])
        if stats["steps"]["pending"]:
            output += colored_comma + pending_word.format(stats["steps"]["pending"])
        output += colored_closing_paren

        if pending_steps:
            sr = StepRegistry()
            pending_step_implementations = make_unique_obj_list(pending_steps, lambda x: x.definition_func)
            output += colorful.white(
                "\nYou have {0} pending step implementation{1} affecting {2} step{3}:\n  {4}\n\nNote: this could be the reason for some failing subsequent steps".format(
                    len(pending_step_implementations),
                    "s" if len(pending_step_implementations) != 1 else "",
                    len(pending_steps),
                    "s" if len(pending_steps) != 1 else "",
                    "\n  ".join(
                        [
                            "-  '{0}' @ {1}".format(
                                sr.get_pattern(s.definition_func),
                                get_func_code(s.definition_func).co_filename,
                            )
                            for s in pending_step_implementations
                        ]
                    ),
                )
            )

        output += "\n"

        if world.config.wip:
            if stats["scenarios"]["passed"] > 0:
                output += colorful.red(
                    "\nThe --wip switch was used, so I didn't expect anything to pass. These scenarios passed:\n"
                )

                has_passed_scenarios = False
                for feature in features:
                    passed_scenarios = list(
                        filter(
                            lambda s: s.state == Step.State.PASSED,
                            feature.all_scenarios,
                        )
                    )
                    for scenario in passed_scenarios:
                        output += colorful.red("\n - {}: {}".format(feature.path, scenario.sentence))
                        has_passed_scenarios = True

                if has_passed_scenarios:
                    output += "\n"
            else:
                output += colorful.green("\nThe --wip switch was used, so the failures were expected. All is good.\n")

        output += colorful.cyan("Run {0} finished within {1}".format(marker, humanize.naturaldelta(duration)))

        write(output)



# -*- coding: utf-8 -*-

"""
    radish
    ~~~~~~

    Behavior Driven Development tool for Python - the root from red to green

    Copyright: MIT, Timo Furrer <tuxtimo@gmail.com>
"""

import pytest

from radish.scenariooutline import ScenarioOutline
from radish.examplescenario import ExampleScenario
from radish.background import Background
from radish.stepmodel import Step
from radish.exceptions import RadishError


def test_creating_simple_scenariooutline():
    """
    Test creating a simple ScenarioOutline
    """
    # given & when
    scenario = ScenarioOutline(
        1,
        "Scenario Outline",
        "Examples",
        "I am a Scenario Outline",
        "foo.feature",
        1,
        parent=None,
        tags=None,
        preconditions=None,
        background=None,
    )

    # then
    assert scenario.id == 1
    assert scenario.keyword == "Scenario Outline"
    assert scenario.example_keyword == "Examples"
    assert scenario.sentence == "I am a Scenario Outline"
    assert scenario.path == "foo.feature"
    assert scenario.line == 1
    assert scenario.parent is None
    assert scenario.tags == []
    assert scenario.preconditions == []
    assert scenario.background is None


def test_building_scenariooutline_scenarios(mocker):
    """
    Test building Scenarios from a Scenario Outline Example
    """
    # given
    scenario_outline = ScenarioOutline(
        1,
        "Scenario Outline",
        "Examples",
        "I am a Scenario Outline",
        "foo.feature",
        1,
        parent=None,
        tags=None,
        preconditions=None,
        background=None,
    )
    # add steps
    scenario_outline.steps.extend(
        [
            mocker.MagicMock(sentence="Given I have <foo>", path="foo.feature"),
            mocker.MagicMock(sentence="And I have <bar>", path="foo.feature"),
            mocker.MagicMock(sentence="When I add those", path="foo.feature"),
        ]
    )
    # add examples
    scenario_outline.examples_header = ["foo", "bar"]
    scenario_outline.examples = [
        # row 0
        ScenarioOutline.Example(["1", "2"], "foo.feature", 1),
        # row 3
        ScenarioOutline.Example(["3", "4"], "foo.feature", 2),
    ]

    # when - build the scenarios
    scenario_outline.build_scenarios()

    # then - expect 2 built Scenarios
    assert len(scenario_outline.scenarios) == 2
    # then - expect that Scenarios are of type ExampleScenario
    assert all(isinstance(x, ExampleScenario) for x in scenario_outline.scenarios)
    # then - expect correct Example Scenario sentences
    assert scenario_outline.scenarios[0].sentence == "I am a Scenario Outline - row 0"
    assert scenario_outline.scenarios[1].sentence == "I am a Scenario Outline - row 1"
    # then - expect correctly replaced Step sentences
    assert scenario_outline.scenarios[0].steps[0].sentence == "Given I have 1"
    assert scenario_outline.scenarios[0].steps[1].sentence == "And I have 2"
    assert scenario_outline.scenarios[0].steps[2].sentence == "When I add those"
    assert scenario_outline.scenarios[1].steps[0].sentence == "Given I have 3"
    assert scenario_outline.scenarios[1].steps[1].sentence == "And I have 4"
    assert scenario_outline.scenarios[1].steps[2].sentence == "When I add those"


def test_building_scenariooutline_scenarios_with_background(mocker):
    """
    Test building Scenarios from a Scenario Outline Example including a Background
    """
    # given
    background = Background("Background", "I am a Background", "foo.feature", 1, parent=None)
    # add some Steps
    background.steps.extend(
        [
            Step(1, "Foo", "foo.feature", 2, background, False),
            Step(2, "Foo", "foo.feature", 3, background, False),
        ]
    )
    scenario_outline = ScenarioOutline(
        1,
        "Scenario Outline",
        "Examples",
        "I am a Scenario Outline",
        "foo.feature",
        1,
        parent=None,
        tags=None,
        preconditions=None,
        background=background,
    )
    # add steps
    scenario_outline.steps.extend(
        [
            mocker.MagicMock(sentence="Given I have <foo>", path="foo.feature"),
            mocker.MagicMock(sentence="And I have <bar>", path="foo.feature"),
            mocker.MagicMock(sentence="When I add those", path="foo.feature"),
        ]
    )
    # add examples
    scenario_outline.examples_header = ["foo", "bar"]
    scenario_outline.examples = [
        # row 0
        ScenarioOutline.Example(["1", "2"], "foo.feature", 1),
        # row 3
        ScenarioOutline.Example(["3", "4"], "foo.feature", 2),
    ]

    # when - build the scenarios
    scenario_outline.build_scenarios()

    # then - expect ExampleScenarios to have background copy assigned
    assert scenario_outline.scenarios[0].background.sentence == "I am a Background"
    assert scenario_outline.scenarios[1].background.sentence == "I am a Background"


def test_scenariooutline_example_colum_width():
    """
    Test calculation for maximum width of Example columns
    """
    # given
    scenario_outline = ScenarioOutline(
        1,
        "Scenario Outline",
        "Examples",
        "I am a Scenario Outline",
        "foo.feature",
        1,
        parent=None,
        tags=None,
        preconditions=None,
        background=None,
    )
    # add examples
    scenario_outline.examples_header = ["foo", "bar"]
    scenario_outline.examples = [
        # row 0
        ScenarioOutline.Example(["Spiderman", "Batman"], "foo.feature", 1),
        # row 3
        ScenarioOutline.Example(["Peter", "Bruce Wayne"], "foo.feature", 2),
    ]

    # when
    index_0_width = scenario_outline.get_column_width(0)
    # then
    assert index_0_width == len("Spiderman")

    # when
    index_1_width = scenario_outline.get_column_width(1)
    # then
    assert index_1_width == len("Bruce Wayne")


def test_scenariooutline_example_invalid_colum_width():
    """
    Test invalid column id exception when calculating column width for Scenario Outline Examples
    """
    # given
    scenario_outline = ScenarioOutline(
        1,
        "Scenario Outline",
        "Examples",
        "I am a Scenario Outline",
        "foo.feature",
        1,
        parent=None,
        tags=None,
        preconditions=None,
        background=None,
    )
    # add examples
    scenario_outline.examples_header = ["foo", "bar"]
    scenario_outline.examples = [
        # row 0
        ScenarioOutline.Example(["Spiderman", "Batman"], "foo.feature", 1),
        # row 3
        ScenarioOutline.Example(["Peter", "Bruce Wayne"], "foo.feature", 2),
    ]

    # when
    with pytest.raises(RadishError) as exc:
        scenario_outline.get_column_width(42)

    # then
    assert str(exc.value) == "Invalid colum_index to get column width for ScenarioOutline 'I am a Scenario Outline'"


def test_scenariooutline_afterparse_logic(mocker):
    """
    Test Scenario Outline after parse logic
    """
    # given
    scenario_outline = ScenarioOutline(
        1,
        "Scenario Outline",
        "Examples",
        "I am a Scenario Outline",
        "foo.feature",
        1,
        parent=None,
        tags=None,
        preconditions=None,
        background=None,
    )
    # add steps
    scenario_outline.steps.extend(
        [
            mocker.MagicMock(sentence="Given I have <foo>", path="foo.feature"),
            mocker.MagicMock(sentence="And I have <bar>", path="foo.feature"),
            mocker.MagicMock(sentence="When I add those", path="foo.feature"),
        ]
    )
    # add examples
    scenario_outline.examples_header = ["foo", "bar"]
    scenario_outline.examples = [
        # row 0
        ScenarioOutline.Example(["1", "2"], "foo.feature", 1),
        # row 3
        ScenarioOutline.Example(["3", "4"], "foo.feature", 2),
    ]

    # when
    scenario_outline.after_parse()

    # then - expect 2 built Scenarios
    assert len(scenario_outline.scenarios) == 2
    assert scenario_outline.complete is True



# -*- coding: utf-8 -*-

"""
    radish
    ~~~~~~

    Behavior Driven Development tool for Python - the root from red to green

    Copyright: MIT, Timo Furrer <tuxtimo@gmail.com>
"""

from datetime import datetime

import pytest
from freezegun import freeze_time

import radish.utils as utils


@pytest.mark.parametrize(
    "basedirs, expected_basedirs, os_name",
    [
        (["foo", "bar"], ["foo", "bar"], "posix"),
        (["foo:bar", "foobar"], ["foo", "bar", "foobar"], "posix"),
        (["foo:bar", "foobar", "one:two:three"], ["foo", "bar", "foobar", "one", "two", "three"], "posix"),
        (["foo:", ":bar"], ["foo", "bar"], "posix"),
        (["C:\\windows\\radish"], ["C:\\windows\\radish"], "nt"),
        (["C:\\windows;radish"], ["C:\\windows", "radish"], "nt"),
    ],
)
def test_flattened_basedirs(mocker, basedirs, expected_basedirs, os_name):
    """
    Test flatten basedirs
    """
    # given & when
    mocker.patch("os.name", os_name)
    actual_basedirs = utils.flattened_basedirs(basedirs)

    # then
    assert actual_basedirs == expected_basedirs


@freeze_time("2015-10-21 04:29:00", tz_offset=+1)
def test_date_time_formatter():
    """
    Test datetime to string format
    """
    # given
    utc_dt = datetime.utcnow()
    expected_datetime_string = "2015-10-21T05:29:00"
    actual_datetime_string = utils.format_utc_to_local_tz(utc_dt)

    # then
    assert actual_datetime_string == expected_datetime_string


def test_make_unique_obj_list():
    """
    Test filter list by propertyName
    """
    object_list = [
        type("SomeObjectClass", (object,), {"propertyName": "1"}),
        type("SomeObjectClass", (object,), {"propertyName": "2"}),
        type("SomeObjectClass", (object,), {"propertyName": "1"}),
    ]

    value_list = utils.make_unique_obj_list(object_list, lambda x: x.propertyName)
    value_list = list(map(lambda x: x.propertyName, value_list))
    value_list.sort()

    assert value_list == ["1", "2"]



# -*- coding: utf-8 -*-

from radish.stepregistry import step
from radish import when, then


@step("I have the number {number:g}")
def have_number(step, number):
    step.context.numbers.append(number)


@when("I divide them")
def sum_numbres(step):
    step.context.result = step.context.numbers[0] / step.context.numbers[1]


@then("I expect the result to be {result:g}")
def expect_result(step, result):
    assert step.context.result == result


@step("I have following numbers")
def have_numbers(step):
    for parameter in step.text.split(","):
        num = int(parameter.split(":")[1])
        step.context.numbers.append(num)


@then("I expect the result")
def expect_result(step):
    result = int(step.text.split(":")[1])
    assert step.context.result == result



# -*- coding: utf-8 -*-

"""
radish
~~~~~~

Behavior Driven Development tool for Python - the root from red to green

Copyright: MIT, Timo Furrer <tuxtimo@gmail.com>
"""

import pytest

from datetime import datetime

from radish.terrain import world
from radish.exceptions import RadishError
from radish.feature import Feature
from radish.scenario import Scenario
from radish.model import Tag
from radish.extensions.junit_xml_writer import JUnitXMLWriter


def test_empty_feature_list():
    writer = JUnitXMLWriter()
    no_features = []

    with pytest.raises(RadishError):
        writer.generate_junit_xml(no_features, "marker-is-ignored")


def test_singel_feature_list(mocker):
    stub = mocker.patch("radish.extensions.junit_xml_writer.JUnitXMLWriter._write_xml_to_disk")

    first_feature = Feature(1, "Feature", "I am a feature", "foo.feature", 1, tags=None)
    first_feature.starttime = datetime.utcnow()
    first_feature.endtime = datetime.utcnow()

    features = [first_feature]

    writer = JUnitXMLWriter()
    writer.generate_junit_xml(features, "marker-is-ignored")

    assert "I am a feature" in str(stub.call_args[0])


def test_normal_feature_list(mocker):
    world.config.junit_relaxed = False
    stub = mocker.patch("radish.extensions.junit_xml_writer.JUnitXMLWriter._write_xml_to_disk")

    tags = [Tag("author", "batman")]

    first_scenario = Scenario(
        1,
        "Scenario",
        "I am a Scenario",
        "foo.feature",
        1,
        parent=None,
        tags=tags,
        preconditions=None,
        background=None,
    )
    first_scenario.starttime = datetime.utcnow()
    first_scenario.endtime = datetime.utcnow()

    first_feature = Feature(1, "Feature", "I am a feature", "foo.feature", 1, tags=None)
    first_feature.starttime = datetime.utcnow()
    first_feature.endtime = datetime.utcnow()
    first_feature.scenarios.append(first_scenario)

    features = [first_feature]

    writer = JUnitXMLWriter()
    writer.generate_junit_xml(features, "marker-is-ignored")

    assert "I am a Scenario" in str(stub.call_args[0])
    assert "properties" not in str(stub.call_args[0])


def test_relaxed_mode_adding_tags_to_junit(mocker):
    world.config.junit_relaxed = True
    stub = mocker.patch("radish.extensions.junit_xml_writer.JUnitXMLWriter._write_xml_to_disk")

    tags = [Tag("author", "batman")]

    first_scenario = Scenario(
        1,
        "Scenario",
        "I am a Scenario",
        "foo.feature",
        1,
        parent=None,
        tags=tags,
        preconditions=None,
        background=None,
    )
    first_scenario.starttime = datetime.utcnow()
    first_scenario.endtime = datetime.utcnow()

    first_feature = Feature(1, "Feature", "I am a feature", "foo.feature", 1, tags=None)
    first_feature.starttime = datetime.utcnow()
    first_feature.endtime = datetime.utcnow()
    first_feature.scenarios.append(first_scenario)

    features = [first_feature]

    writer = JUnitXMLWriter()
    writer.generate_junit_xml(features, "marker-is-ignored")

    assert "author" in str(stub.call_args[0])
    assert "batman" in str(stub.call_args[0])



# -*- coding: utf-8 -*-

import re

from radish.stepregistry import step
from radish import given, when, then


# @step(re.compile(r"I have the number (\d+)"))
@step("I have the number {number:g}")
def have_number(step, number):
    step.context.numbers.append(int(number))


@when("I sum them")
def sum_numbers(step):
    step.context.result = sum(step.context.numbers)


@then("I expect the result to be {result:g}")
def expect_result(step, result):
    assert step.context.result == result



# -*- coding: utf-8 -*-

"""
    radish
    ~~~~~~

    Behavior Driven Development tool for Python - the root from red to green

    Copyright: MIT, Timo Furrer <tuxtimo@gmail.com>
"""

import pytest

from radish.runner import Runner
from radish.stepmodel import Step


@pytest.mark.parametrize(
    "run_state, expected_returncode",
    [
        (Step.State.PASSED, 0),
        (Step.State.UNTESTED, 0),
        (Step.State.SKIPPED, 0),
        (Step.State.PENDING, 0),
        (Step.State.FAILED, 1),
    ],
    ids=[
        "Running Step returing PASSED",
        "Running Step returing UNTESTED",
        "Running Step returing SKIPPED",
        "Running Step returing PENDING",
        "Running Step returing FAILED",
    ],
)
def test_run_single_step(run_state, expected_returncode, hookregistry, mocker):
    """
    Test running a single Step
    """
    # given
    runner = Runner(hookregistry)
    step = mocker.MagicMock()
    step.run.return_value = run_state

    # when
    returncode = runner.run_step(step)

    # then
    assert returncode == expected_returncode
    assert step.run.call_count == 1


@pytest.mark.parametrize(
    "debug_state, expected_returncode",
    [
        (Step.State.PASSED, 0),
        (Step.State.UNTESTED, 0),
        (Step.State.SKIPPED, 0),
        (Step.State.PENDING, 0),
        (Step.State.FAILED, 1),
    ],
    ids=[
        "Debugging Step returing PASSED",
        "Debugging Step returing UNTESTED",
        "Debugging Step returing SKIPPED",
        "Debugging Step returing PENDING",
        "Debugging Step returing FAILED",
    ],
)
def test_debug_single_step(debug_state, expected_returncode, world_config, hookregistry, mocker):
    """
    Test debugging a single Step
    """
    # given
    runner = Runner(hookregistry)
    step = mocker.MagicMock()
    step.debug.return_value = debug_state

    # set debug mode
    world_config.debug_steps = True

    # when
    returncode = runner.run_step(step)

    # then
    assert returncode == expected_returncode
    assert step.debug.call_count == 1


def test_run_single_step_show_only(hookregistry, mocker):
    """
    Test running single step when show only mode is on
    """
    # given
    runner = Runner(hookregistry, show_only=True)
    step = mocker.MagicMock()
    step.run.return_value = Step.State.FAILED

    # when
    returncode = runner.run_step(step)

    # then
    assert returncode == 0
    assert step.run.call_count == 0


def test_skip_single_step(hookregistry, mocker):
    """
    Test skipping a single Step
    """
    # given
    runner = Runner(hookregistry)
    step = mocker.MagicMock()
    step.skip.return_value = None

    # when
    runner.skip_step(step)

    # then
    assert step.skip.call_count == 1


@pytest.mark.parametrize("run_or_skip", ["run_step", "skip_step"])
def test_run_skip_step_hooks(run_or_skip, hookregistry, mocker):
    """
    Test that Hooks are executed when running or skipping a Step
    """
    # given
    # register hooks in registry
    before_step_stub = mocker.stub()
    after_step_stub = mocker.stub()
    hookregistry.register("before", "each_step", before_step_stub)
    hookregistry.register("after", "each_step", after_step_stub)
    # create runner
    runner = Runner(hookregistry, show_only=True)

    step = mocker.MagicMock(all_tags=[])

    # when
    method = getattr(runner, run_or_skip)
    method(step)

    # then
    assert before_step_stub.call_count == 1
    assert after_step_stub.call_count == 1


def test_should_call_hooks_in_correct_order(hookregistry, mocker):
    """
    Test that hooks are called in correct order.

    Correct order meaning:
        * before hooks: ascending order
        * after hooks: descending order
    """
    # given
    runner = Runner(hookregistry)

    # register hooks
    data = []

    def first_before_stub(*args):
        data.append(1)

    def first_after_stub(*args):
        data.append(1)

    def second_before_stub(*args):
        data.append(2)

    def second_after_stub(*args):
        data.append(2)

    hookregistry.register("before", "each_step", second_before_stub, order=2)
    hookregistry.register("after", "each_step", second_after_stub, order=2)
    hookregistry.register("before", "each_step", first_before_stub, order=1)
    hookregistry.register("after", "each_step", first_after_stub, order=1)

    # setup dummy step
    step = mocker.MagicMock()

    # when
    runner.run_step(step)

    # then
    assert data == [1, 2, 2, 1]



# -*- coding: utf-8 -*-

"""
This module provide Python 3 compatability
"""

import sys
import re

try:
    re_pattern = re.Pattern  # >= Python 3.7
except AttributeError:
    re_pattern = re._pattern_type


# RecursionError does not exist in Python < 3.5
RecursionError = RuntimeError if sys.version_info < (3, 5) else RecursionError



# -*- coding: utf-8 -*-

"""
This module contains a class to load the step and terrain files
"""

import os
import sys
import fnmatch

from . import utils


def load_modules(location):
    """
    Loads all modules in the `location` folder
    """
    if os.name == "nt":
        location = location.replace("$PWD", os.getcwd())

    location = os.path.expanduser(os.path.expandvars(location))
    if not os.path.exists(location):
        raise OSError("Location '{0}' to load modules does not exist".format(location))

    for p, _, f in os.walk(location):
        for filename in fnmatch.filter(f, "*.py"):
            load_module(os.path.join(p, filename))


def load_module(path):
    """
    Loads a module by the given `path`

    :param string path: the path to the module to load
    """
    module_name = os.path.splitext(os.path.split(path)[1])[0]
    try:
        import importlib.util

        spec = importlib.util.spec_from_file_location(module_name, path)
        module = importlib.util.module_from_spec(spec)
        spec.loader.exec_module(module)
    except Exception as e:
        raise ImportError("Unable to import module '{0}' from '{1}': {2}".format(module_name, path, e))



# -*- coding: utf-8 -*-

"""
This module provides functionaliy to test
if some sentences are matched with the expected
step implementations.
"""

import sys
import codecs

import yaml
import colorful

from radish.loader import load_modules
from radish.matcher import match_step
from radish.stepregistry import StepRegistry
from radish.utils import get_func_arg_names, get_func_location, locate


def test_step_matches_configs(match_config_files, basedirs, cover_min_percentage=None, cover_show_missing=False):
    """
    Test if the given match config files matches the actual
    matched step implementations.
    """
    if cover_min_percentage is not None and float(cover_min_percentage) > 100:
        sys.stderr.write(
            str(
                colorful.magenta(
                    "You are a little cocky to think you can reach a minimum coverage of {0:.2f}%\n".format(
                        float(cover_min_percentage)
                    )
                )
            )
        )
        return 3

    # load user's custom python files
    for basedir in basedirs:
        load_modules(basedir)

    steps = StepRegistry().steps

    if not steps:
        sys.stderr.write(
            str(
                colorful.magenta(
                    "No step implementations found in {0}, thus doesn't make sense to continue".format(basedirs)
                )
            )
        )
        return 4

    failed = 0
    passed = 0
    covered_steps = set()

    for match_config_file in match_config_files:
        # load the given match config file
        with codecs.open(match_config_file, "r", "utf-8") as f:
            match_config = yaml.safe_load(f)

        if not match_config:
            print(colorful.magenta("No sentences found in {0} to test against".format(match_config_file)))
            return 5

        print(colorful.yellow("Testing sentences from {0}:".format(colorful.bold_yellow(match_config_file))))
        failed_sentences, passed_senteces = test_step_matches(match_config, steps)
        failed += failed_sentences
        passed += passed_senteces

        covered_steps = covered_steps.union(x["should_match"] for x in match_config if "should_match" in x)

        # newline
        sys.stdout.write("\n")

    report = colorful.bold_white("{0} sentences (".format(failed + passed))
    if passed > 0:
        report += colorful.bold_green("{0} passed".format(passed))

    if passed > 0 and failed > 0:
        report += colorful.bold_white(", ")

    if failed > 0:
        report += colorful.bold_red("{0} failed".format(failed))
    report += colorful.bold_white(")")
    print(report)

    step_coverage = 100.0 / len(steps) * len(covered_steps)
    coverage_report = colorful.bold_white(
        "Covered {0} of {1} step implementations".format(len(covered_steps), len(steps))
    )

    ret = 0 if failed == 0 else 1

    if cover_min_percentage:
        coverage_color = colorful.bold_green if step_coverage >= float(cover_min_percentage) else colorful.bold_red
        coverage_report += colorful.bold_white(" (coverage: ")
        coverage_report += coverage_color("{0:.2f}%".format(step_coverage))
        if float(cover_min_percentage) > step_coverage:
            coverage_report += colorful.bold_white(
                ", expected a minimum of {0}".format(colorful.bold_green(cover_min_percentage + "%"))
            )
            if failed == 0:
                ret = 2
            # if tests have passed and coverage is too low we fail with exit code 2
        coverage_report += colorful.bold_white(")")

    print(coverage_report)

    if cover_show_missing:
        missing_steps = get_missing_steps(steps, covered_steps)
        if missing_steps:
            missing_step_report = colorful.bold_yellow("Missing steps:\n")
            for step in missing_steps:
                missing_step_report += "- {0} at ".format(colorful.cyan(step[0]))
                missing_step_report += colorful.cyan(step[1]) + "\n"
            sys.stdout.write(str(missing_step_report))

    return ret


def test_step_matches(match_config, steps):
    """
    Test if the given match config matches the actual
    matched step implementations.
    """
    failed = 0
    passed = 0

    for item in match_config:
        validate_config_item(item)

        sentence = item["sentence"]

        if "should_match" in item:
            has_passed = test_step_match(sentence, item["should_match"], item.get("with_arguments", None), steps)
        else:
            has_passed = test_step_not_match(sentence, item["should_not_match"], steps)

        if has_passed:
            passed += 1
        else:
            failed += 1

    return failed, passed


def test_step_match(sentence, expected_step, expected_arguments, steps):
    sys.stdout.write(
        '{0} STEP "{1}" SHOULD MATCH {2}    '.format(
            colorful.yellow(">>"), colorful.cyan(sentence), colorful.cyan(expected_step)
        )
    )

    result = match_step(sentence, steps)
    if not result:
        output_failure(None, ["Expected sentence didn't match any step implementation"])
        return False

    if expected_step != result.func.__name__:
        output_failure(
            result.func,
            ["Expected sentence matched {0} instead of {1}".format(result.func.__name__, expected_step)],
        )
        return False

    if expected_arguments:
        arguments = merge_step_args(result)
        expected_arguments = {k: v for expected_arguments in expected_arguments for k, v in expected_arguments.items()}
        argument_errors = check_step_arguments(expected_arguments, arguments)
        if argument_errors:
            output_failure(result.func, argument_errors)
            return False

    print(str(colorful.bold_green("✔")))
    return True


def test_step_not_match(sentence, expected_not_matching_step, steps):
    step_to_print = colorful.cyan(expected_not_matching_step) if expected_not_matching_step else "ANY"
    sys.stdout.write(
        '{0} STEP "{1}" SHOULD NOT MATCH {2}    '.format(colorful.yellow(">>"), colorful.cyan(sentence), step_to_print)
    )

    result = match_step(sentence, steps)
    if result:
        if not expected_not_matching_step or result.func.__name__ == expected_not_matching_step:
            output_failure(
                None,
                ["Expected sentence did match {0} but it shouldn't".format(expected_not_matching_step)],
            )
            return False

    print(str(colorful.bold_green("✔")))
    return True


VALID_CONFIG_ITEMS = {"sentence", "should_match", "should_not_match", "with_arguments"}


def validate_config_item(config):
    """
    Validate the given config object
    """
    given_attributes = set(config.keys())
    if not given_attributes.issubset(VALID_CONFIG_ITEMS):
        raise ValueError(
            "The config attributes {0} are invalid. Use only {1}".format(
                ", ".join(sorted(given_attributes.difference(VALID_CONFIG_ITEMS))),
                ", ".join(sorted(VALID_CONFIG_ITEMS)),
            )
        )

    if "sentence" not in config or ("should_match" not in config and "should_not_match" not in config):
        raise ValueError(
            "You have to provide a sentence and the function name which should (not) be matched (should_match, should_not_match)"
        )


def output_failure(step_func, errors):
    """
    Write the given errors to stdout.
    """
    sys.stdout.write(str(colorful.bold_red("✘")))
    if step_func is not None:
        sys.stdout.write(str(colorful.red(" (at {0})".format(get_func_location(step_func)))))

    sys.stdout.write("\n")

    for error in errors:
        print(str(colorful.red("  - {0}".format(error))))


def check_step_arguments(expected_arguments, arguments):
    """
    Check if the given expected arguments
    match the actual arguments
    """
    errors = []
    for arg_name, arg_value in expected_arguments.items():
        if arg_name not in arguments:
            errors.append(
                'Expected argument "{0}" is not in matched arguments {1}'.format(arg_name, list(arguments.keys()))
            )
            continue

        use_repr = False

        # check if argument value is a dict, if yes we'll do thorough comparison
        if isinstance(arg_value, dict) and "type" not in arg_value.keys():
            _type = "dict"
            value = arg_value
        elif isinstance(arg_value, dict):
            _type = arg_value["type"]
            value = arg_value["value"]
            # Use repr protocol to match argument values
            use_repr = "use_repr" in arg_value and arg_value["use_repr"]

            # check if value should be casted to the given type
            if "cast" in arg_value and arg_value["cast"] is True:
                obj_type = locate(_type)
                if obj_type is None:
                    errors.append('Cannot cast to type "{0}" because it is unknown'.format(_type))
                    continue

                try:
                    value = obj_type(value)
                except Exception as exc:
                    errors.append('Failed to cast "{0}" to given type "{1}"'.format(value, type))
                    continue
        else:
            _type = type(arg_value).__name__
            value = arg_value

        if not use_repr and _type != type(value).__name__:
            errors.append(
                'Conflicting argument configuration: given value is actually of type "{0}" although it should match a value of type "{1}"'.format(
                    type(value).__name__, _type
                )
            )
            continue

        if type(arguments[arg_name]).__name__ != _type:
            errors.append(
                'Expected argument "{0}" is of type "{1}" instead "{2}"'.format(
                    arg_name, type(arguments[arg_name]).__name__, _type
                )
            )
            continue

        matched = None
        if use_repr:
            matched = repr(arguments[arg_name]) != value
        else:
            matched = arguments[arg_name] != value

        if matched:
            errors.append(
                'Expected argument "{0}" with value "{1}" does not match value "{2}"'.format(
                    arg_name, value, arguments[arg_name]
                )
            )
    return errors


def merge_step_args(step_func):
    """
    Merges the arguments and keyword arguments
    of the given step function.
    """
    #: Holds the merged arguments as a dict with the corresponding matched values
    args, kwargs = step_func.argument_match.evaluate()
    step_arg_names = get_func_arg_names(step_func.func)[1:]
    arguments = dict(zip(step_arg_names, args))
    arguments.update(kwargs)
    return arguments


def get_missing_steps(steps, covered_steps):
    """
    Get all steps within ``steps`` which are not
    covered by ``covered_steps``.
    """
    missing_steps = []
    for step_func in steps.values():
        if step_func.__name__ not in covered_steps:
            missing_steps.append((step_func.__name__, get_func_location(step_func)))
    return missing_steps



from radish import given, when, then


@when("I have the value {:Boolean}")
def have_value(step, boolean):
    step.context.boolean = boolean


@then("I expect it to be parsed as True")
def expect_true(step):
    assert step.context.boolean is True


@then("I expect it to be parsed as False")
def expect_true(step):
    assert step.context.boolean is False



# -*- coding: utf-8 -*-

from radish import given, when, then
from radish import world

from example.model import Hero


@given("I setup the hero database")
def setup_hero_db(step):
    pass


@when("I query for the hero with name {hero:Hero}")
def query_hero(step, hero):
    step.context.last_queried_hero = hero


@then("I expect it's name to be {forename:w} {surname:w}")
def expect_hero(step, forename, surname):
    assert step.context.last_queried_hero.forename == forename
    assert step.context.last_queried_hero.surname == surname
